from sigla.core.outputs.OutputFile import OutputFile
from sigla.utils import import_node

Node = import_node()


class NodeFile(Node):
    def process(self):
        super().process()

        content = self.children()

        if "to" not in self.data.attributes.keys():
            self.raise_missing_to_attribute()

        to = self.data.attributes["to"]

        return OutputFile(to, content)

    def raise_missing_to_attribute(self):
        raise AttributeError(
            "You need to provide the propriety `to` with a filepath on the element <file> to save the results to"
            # noqa
        )



from dataclasses import dataclass, field


@dataclass
class Data:
    """This data class is meant to hold the data for each node"""

    tag: str
    attributes: dict = field(default_factory=dict)
    frontmatter_attributes: dict = field(default_factory=dict)
    children: list = field(default_factory=list)
    parent_attributes: dict = field(default_factory=dict)



class TemplateDoesNotExistError(Exception):
    def __init__(self, tag, from_entity, *args: object) -> None:
        self.message = f"Missing template {tag}"
        self.from_entity = from_entity
        super().__init__(self.message, *args)



from pathlib import Path
from typing import Union

from sigla.utils import ensure_dirs


class OutputFile:
    path: str = None
    content: str = None

    def __init__(self, path, content):
        self.path = path
        self.content = content

    def save(self):
        ensure_dirs(Path(self.path).parent)
        with open(self.path, "w") as h:
            h.write(self.content)

    def __eq__(self, o: Union[object, "OutputFile"]) -> bool:
        if type(self) != type(o):
            return False
        if self.content != o.content:
            return False
        if self.path != o.path:
            return False
        return True

    def __repr__(self):
        return f"[file-output: {self.content}]"



from jinja2 import Environment, BaseLoader

__RENDERS = dict()


def register(f):
    __RENDERS[f.__name__] = f
    return f


def render(plugin, template, **kw):
    return __RENDERS[plugin](template, **kw)


@register
def jinja(template: str, filters: dict, **kwargs: any) -> str:
    env = Environment(loader=BaseLoader)
    env.filters.update(filters)
    template = env.from_string(template)
    return template.render(**kwargs)



import textwrap
from pathlib import Path

from sigla import config
from sigla.utils import ensure_dirs


def get_content():
    return textwrap.dedent(
        """
        \"\"\"
        Export filters to use on the templates using the `FILTERS` variable
        \"\"\"
        import json
        from sigla.filters import *

        @register_filter('dump')
        def dump(var):
            return json.dumps(var, indent=4)

        """
    )


def init_command():
    print(":: sigla init")
    print(f":: - checking/creating folder {config.path.templates}")
    print(f":: - checking/creating folder {config.path.snapshots}")
    print(f":: - checking/creating folder {config.path.definitions}")

    ensure_dirs(
        config.path.templates,
        config.path.snapshots,
        config.path.definitions,
    )

    print(f":: - checking/creating file {config.path.filters}")
    filters_path = Path(config.path.filters)
    filters_file_exists = filters_path.exists()
    if not filters_file_exists:
        filters_path.write_text(get_content())



import inspect


class NodeList(list):
    methods = [
        "__call__",
        "flatten",
        "uniq",
        "nonone",
        "get",
        "join",
    ]

    def __call__(self, sep="\n"):
        children = [c() for c in self]
        if sep:
            return sep.join(children)
        return children

    def flatten(self):
        ret = []
        for child in self:
            if not child:
                continue
            if hasattr(child, "flatten"):
                ret += child.flatten()
            elif isinstance(child, list):
                ret += child
            else:
                raise Exception(f"Case not accounted for: {type(child)}")
        return self.__class__(ret)

    def uniq(self):
        ret = self.__class__()
        for child in self:
            if child not in ret:
                ret.append(child)
        return ret

    def nonone(self):
        return self.__class__(filter(lambda e: e is not None, self))

    def get(self, name):
        return self.__class__(map(lambda e: getattr(e, name, None), self))

    def join(self, sep):
        return sep.join(self)

    @classmethod
    def get_node_list_methods(cls):
        methods_children = []
        for method in cls.methods:
            args = inspect.getfullargspec(getattr(NodeList, method)).args
            args = [a for a in args if a not in ["self"]]
            method = f".{method}" if method != "__call__" else ""
            methods_children.append(
                f"node.children{method}({','.join(args)})"
            )
        return methods_children



from pathlib import Path
from typing import Union

import typer

from sigla import config, from_xml_string
from sigla.core.outputs.OutputEcho import OutputEcho
from sigla.core.outputs.OutputFile import OutputFile
from sigla.errors import TemplateDoesNotExistError
import abc


class AbstractDefinitionBasedCommand(abc.ABC, object):
    def __init__(self, references):
        self.references = references
        self.globs = [
            Path(config.path.definitions).glob(f"{reference}.xml")
            for reference in references
        ]
        self.matches = [match for glob in self.globs for match in glob]

    @abc.abstractmethod
    def __call__(self, *args, **kwargs):
        pass


class RunCommand(AbstractDefinitionBasedCommand):
    @staticmethod
    def handle_results(s: Union[OutputFile, OutputEcho]):
        if isinstance(s, OutputFile):
            print(f":: Saving {s.path}")
            s.save()

    def handle_match(self, match):
        if not match.exists():
            raise typer.Exit(f"✋ The definition(s) do not exists {match}")

        is_dir = match.is_dir()

        if is_dir:
            return

        print(f":: Reading {match}")

        str_xml = match.read_text()
        results = from_xml_string(str_xml)()

        for output in results:
            self.handle_results(output)

    def __call__(self, *args, **kwargs):

        if len(self.matches) == 0:
            try:
                print(f"✋ No definition(s) found for {self.references}")
            except TemplateDoesNotExistError as e:
                print(e)
                raise typer.Exit(e)

        for match in self.matches:
            self.handle_match(match)



from typing import Union
from sigla.core.nodes.NodeEcho import NodeEcho
from sigla.core.nodes.NodeFile import NodeFile
from sigla.core.nodes.NodeRoot import NodeRoot
from sigla.core.nodes.NodeTemplate import NodeTemplate

NodeType = Union[NodeRoot, NodeFile, NodeEcho, NodeTemplate]



from typing import Union


class OutputEcho:
    content: str = None

    def __init__(self, content):
        self.content = content

    def __eq__(self, o: Union[object, "OutputEcho"]) -> bool:
        if type(self) != type(o):
            return False
        if self.content != o.content:
            return False
        return True

    def __repr__(self):
        return f"[echo-output: {self.content}]"



import environ


@environ.config(prefix="")
class Config:
    @environ.config
    class Paths:
        templates = environ.var(".sigla/templates")
        snapshots = environ.var(".sigla/snapshots")
        definitions = environ.var(".sigla/definitions")
        filters = environ.var(".sigla/filters.py")

    @environ.config
    class Cls:
        node_list = environ.var("sigla.core.cls.NodeList.NodeList")
        node = environ.var("sigla.core.cls.Node.Node")

    path: Paths = environ.group(Paths)
    cls: Cls = environ.group(Cls)


config = environ.to_config(Config)



from sigla.core.outputs.OutputEcho import OutputEcho
from sigla.utils import import_node

Node = import_node()


class NodeEcho(Node):
    def process(self):
        super().process()
        content = self.children()
        return OutputEcho(content)



import json
import types
import warnings
import importlib
import importlib.machinery
from contextlib import suppress
from pathlib import Path
from textwrap import dedent

from frontmatter import u, detect_format, handlers
from yaml.parser import ParserError

#
from sigla import config


def ensure_dirs(*paths: str):
    """Ensure the paths exist"""
    for path_ in paths:
        Path(path_).mkdir(parents=True, exist_ok=True)


def import_reference(ref: str) -> type:
    """Imports stuff from modules dinamically"""
    p, m = ref.rsplit(".", 1)
    mod = importlib.import_module(p)
    res = getattr(mod, m)
    return res


def import_node_list() -> type:
    """Imports NodeList"""
    ref = config.cls.node_list
    res = import_reference(ref)
    return res


def import_node() -> type:
    """Imports NodeList"""
    ref = config.cls.node
    res = import_reference(ref)
    return res


def cast_property(prop_name, prop_value):
    if prop_name.endswith("-int"):
        return prop_name.replace("-int", ""), int(prop_value)

    elif prop_name.endswith("-float"):
        new_key = prop_name.replace("-float", "")
        new_value = float(prop_value)
        return new_key, new_value

    elif prop_name.endswith("-json"):
        new_key = prop_name.replace("-json", "")
        new_value = json.loads(prop_value)
        return new_key, new_value

    elif prop_name.endswith("-bool"):
        new_key = prop_name.replace("-bool", "")
        new_value = True if prop_value in ["True", "true", "1"] else False
        return new_key, new_value

    new_key = prop_name
    new_value = prop_value
    return new_key, new_value




def cast_xml_property(prop_name, prop_value):
    warnings.warn("This will be deprecated. Please us `cast_property`")
    return cast_property(prop_name, prop_value)


def load_module(module_name, module_path):
    loader = importlib.machinery.SourceFileLoader(module_name, module_path)
    module = types.ModuleType(loader.name)
    loader.exec_module(module)
    return module


def load_filters_from(module_path):
    filters = {}
    module_name = "filters"
    with suppress(FileNotFoundError):
        filters_module = load_module(module_name, module_path)
        if "FILTERS" in dir(filters_module):
            filters = filters_module.FILTERS
    return filters


def frontmatter_split(text, *, encoding="utf-8", handler=None):
    text = u(text, encoding).strip()

    # this will only run if a handler hasn't been set higher up
    handler = handler or detect_format(text, handlers)
    if handler is None:
        return None, text, None

    try:
        fm, content = handler.split(text)
    except ValueError:
        return None, text, handler

    return fm, content, handler


def frontmatter_parse(fm, handler, *, metadata=None):
    if metadata is None:
        metadata = {}

    try:
        fm = handler.load(fm)
    except ParserError as e:
        print(
            dedent(
                f"""
        ===
        There is an error on the following yaml (front matter)

        {fm}

        ===

        """
            )
        )
        raise e

    if isinstance(fm, dict):
        metadata.update(fm)

    return metadata


def get_template_path(base_path, tag, ext="jinja2", bundle=None):
    path = Path(base_path)
    if bundle:
        path = path.joinpath(bundle)
    path = path.joinpath(f"{tag}.{ext}")
    return path





from pathlib import Path
from xml.etree import ElementTree
from sigla.core.cls.Data import Data
from sigla.core.nodes.NodeEcho import NodeEcho
from sigla.core.nodes.NodeFile import NodeFile
from sigla.core.nodes.NodeRoot import NodeRoot
from sigla.core.nodes.NodeTemplate import NodeTemplate
from sigla.types import NodeType
from sigla.utils import cast_property


def node_factory(tag, attributes) -> NodeType:
    if tag == "root":
        return NodeRoot(tag, attributes=attributes)
    elif tag == "file":
        return NodeFile(tag, attributes=attributes)
    elif tag == "echo":
        return NodeEcho(tag, attributes=attributes)
    else:
        return NodeTemplate(tag, attributes=attributes)


def data_to_node(node: Data, *, factory=node_factory) -> NodeType:
    """Takes Data classes and converts them to Node"""
    ret = factory(node.tag, node.attributes)

    for r in node.children:
        child = data_to_node(r, factory=factory)
        ret.append(child)

    return ret


def xml_to_data(obj) -> Data:
    attributes = obj.attrib.copy()
    new_attributes = {}

    for prop_name, prop_value in attributes.items():
        prop_name, prop_value = cast_property(prop_name, prop_value)
        new_attributes[prop_name] = prop_value

    node = Data(obj.tag, attributes=new_attributes, children=[])

    for child in obj:
        node.children.append(xml_to_data(child))

    return node


def string_to_data(s) -> Data:
    obj: ElementTree = ElementTree.fromstring(s)
    node = xml_to_data(obj)
    return node


def from_xml_string(source, *, factory=node_factory) -> NodeType:
    return data_to_node(string_to_data(source), factory=factory)


def from_xml_file(source, *, factory=node_factory) -> NodeType:
    source = Path(source).read_text()
    return from_xml_string(source, factory=factory)



import textwrap
from pathlib import Path

import typer

from sigla import config


def get_content(name):
    return textwrap.dedent(
        f"""\
        <root>
            <file to="output/{name}[.ext]">
                <{name}>
                    [...]
                </{name}>
            </file>
        </root>
    """
    )


def new_command(name):
    print(f":: creating new definition: {name}")

    destination = Path(config.path.definitions).joinpath(f"{name}.xml")

    if destination.exists():
        raise typer.Exit("✋ This definition already exists")

    destination.write_text(get_content(name))



import typer
from typing import List
from sigla import __version__
from sigla.cli.commands.RunCommand import RunCommand
from sigla.cli.commands.dump_config_command import dump_config_command
from sigla.cli.commands.init_command import init_command
from sigla.cli.commands.new_command import new_command

app = typer.Typer()


@app.command()
def dump_config():
    dump_config_command()


@app.command()
def init():
    """ Creates the .init folder to keep stuff """
    init_command()


@app.command()
def new(name: str):
    """ Generate a new definition for a generator. """
    new_command(name)


@app.command()
def run(references: List[str]):
    """ run a generator """
    run_command = RunCommand(references)
    run_command()


@app.command()
def version():
    """ Print the version """
    print(f"Version: {__version__}")


if __name__ == "__main__":
    app()

# class SnapshotCommand(AbstractDefinitionBasedCommand):
#     pass
#
#
# @app.command()
# def snapshot(references: List[str]):
#     """
#     Make snapshots
#     """
#     snapshot_command = SnapshotCommand(references)
#     snapshot_command()
#
#
# @cli.command()
# @click.argument("files", nargs=-1, type=click.Path(exists=True))
# def snapshot(files):
#     """
#     Makes snapshots
#     """
#     for file in files:
#         cli_make_snapshots(cli_read_snapshot_definition(file))
#
#
# @cli.command()
# @click.argument("files", nargs=-1, type=click.Path(exists=True))
# def verify(files):
#     """
#     Verify snapshots
#     """
#     for file in files:
#         cli_verify_snapshots(cli_read_snapshot_definition(file))
#
#
# <ut kgdir="kg">
#   <test cmd="ruby -I../lp rpcgen.rb examples/Test.java">
#     <out>examples/TestHandler.java</out>
#     <out>examples/TestStubs.java</out>
#   </test>
# </ut>
#
#
#
# def cli_read_snapshot_definition(file):
#     print(f":: Loading {file}")
#     doc = load_xml_from_file(file)
#     print(f":: Ensure {SNAPSHOTS_DIRECTORY} exists")
#     ensure_dirs(SNAPSHOTS_DIRECTORY)
#     print(":: Reading tests")
#     tests = []
#     for test_node in doc.iter("test"):
#         files = [test_node.attrib["out"]]
#
#         for out_node in test_node.iter("out"):
#             out = out_node.text.strip()
#             files.append(out)
#
#         tests.append(
#             {"command": test_node.attrib["cmd"], "output_files": files}
#         )
#     print(f"    ‣ Loaded {len(tests)} commands")
#     return tests
#
#
# def cli_make_snapshots(tests):
#     print(":: Making snapshots")
#     for test in tests:
#         print(f'    ‣ Command {test["command"]}')
#         os.system(test["command"])
#
#         for file in test["output_files"]:
#             gn = SNAPSHOTS_DIRECTORY + "/" + file
#
#             #
#             # MAKING
#             #
#             ensure_parent_dir(gn)
#             print(f"        ‣ Saving snapshot {file} to {gn}")
#             copyfile(file, gn)
#
#
# def cli_verify_snapshots(tests):
#     print(":: Checking snapshots")
#     failures = []
#     for test in tests:
#         print(f'    ‣ Command {test["command"]}')
#         os.system(test["command"])
#
#         for file in test["output_files"]:
#             gn = SNAPSHOTS_DIRECTORY + "/" + file
#
#             with open(file, "r") as h:
#                 current_result = h.read()
#
#             #
#             # TESTING
#             #
#             print(f"        ‣ Checking snapshot {file} against {gn}")
#             with open(gn, "r") as h:
#                 good_result = h.read()
#
#             if good_result != current_result:
#                 print("        🚩 Snapshot comparison failed")
#                 failures.append(test["command"])



from collections import ChainMap
from copy import deepcopy


class Attributes(ChainMap):
    def without(self, *names):
        data = deepcopy(dict(self))
        for name in names:
            del data[name]
        return Attributes(data)

    def as_kwargs(self, sep=","):
        kwargs = []
        for k, v in self.items():
            if type(v) == int:
                kwargs.append(f"{k}={v}")
            else:
                kwargs.append(f'{k}="{v}"')
        if sep:
            return ", ".join(kwargs)
        return kwargs



from textwrap import dedent
from typing import Any
from sigla.core.cls.Attributes import Attributes
from sigla.core.cls.Data import Data
from sigla.utils import import_node_list

NodeList = import_node_list()


class Node(object):
    def __init__(self, tag, attributes=None):
        if attributes is None:
            attributes = {}
        self.data = Data(
            tag, attributes=attributes, children=[], parent_attributes={}
        )
        self.context = {}

    @property
    def attributes(self):
        return Attributes(
            self.data.frontmatter_attributes,
            self.data.attributes,
            self.data.parent_attributes,
        )

    def __getattr__(self, name: str) -> Any:
        if name in self.attributes:
            return self.attributes[name]
        raise AttributeError(f"{self.__class__.__name__}.{name} is invalid.")

    def __repr__(self) -> str:
        tag = self.data.tag
        attrs = self.attributes if len(self.attributes.keys()) > 0 else ""
        children = self.children if len(self.children) > 0 else ""
        return dedent(f"""<{tag} {attrs}>{children}</{self.data.tag}>""")

    def __eq__(self, other):
        return self.data == other.data

    def append(self, node: "Node"):
        self.data.children.append(node)
        return self

    def process(self):
        self.update_context()

    def __call__(self, *args, **kwargs):
        return self.process()

    def update_parent_context(self, ctx):
        self.data.parent_attributes.update(**ctx)

    def update_context(self):
        for child in self.data.children:
            ctx = self.context.copy()
            ctx.update(dict(self.attributes))
            child.update_parent_context(ctx)
            child.update_context()

    def flatten(self):
        return [self, *self.children.flatten()]

    @property
    def children(self):
        return NodeList(self.data.children)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['sigla',
 'sigla.cli',
 'sigla.cli.commands',
 'sigla.conf',
 'sigla.core',
 'sigla.core.cls',
 'sigla.core.nodes',
 'sigla.core.outputs']

package_data = \
{'': ['*']}

install_requires = \
['Jinja2>=2.11.2,<3.0.0',
 'configparser>=5.0.1,<6.0.0',
 'environ-config>=20.1.0,<21.0.0',
 'pretty-errors>=1.2.19,<2.0.0',
 'pydash>=4.9.1,<5.0.0',
 'python-frontmatter>=0.5.0,<0.6.0',
 'shellingham>=1.4.0,<2.0.0',
 'typer>=0.3.2,<0.4.0']

entry_points = \
{'console_scripts': ['sigla = sigla.cli.cli:app']}

setup_kwargs = {
    'name': 'sigla',
    'version': '0.0.59',
    'description': 'Yet another code generator',
    'long_description': None,
    'author': 'mg santos',
    'author_email': 'mauro.goncalo@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from sigla import config


def dump_config_command():
    print(f"|> config.path.templates:\t {config.path.templates}")
    print(f"|> config.path.snapshots:\t {config.path.snapshots}")
    print(f"|> config.path.definitions:\t {config.path.definitions}")
    print(f"|> config.path.filters:\t {config.path.filters}")
    print("")
    print(f"|> config.cls.node_list:\t {config.cls.node_list}")



from textwrap import dedent

from jinja2 import (
    UndefinedError,
)

from sigla import config
from sigla.core.renders import render
from sigla.utils import (
    frontmatter_split,
    frontmatter_parse,
    import_node,
    import_node_list,
    load_filters_from, get_template_path)

Node = import_node()
NodeList = import_node_list()


def get_default_error_message(node, str_tpl):
    return dedent(
        f"""\
        ------------------------------
        ERROR WHILE RENDERING TEMPLATE
        ------------------------------

        TEMPLATE:
        {str_tpl}

        NODE:
        {node}

        ---

        """
    )


default_template_content = dedent(
    """\
    ---
    title: Generating code like it's 99
    ---

    You can use variables written inside the frontmatter:
        - {% raw %}{{ node.title }}{% endraw %}: Generating code like it's 99

    Available variables (with current values for first node as an example):
    {%- for key, value in variables.items() %}
        - {{ "{{ node." + key + " }}" }}: {{value}}
    {%- else %}
        - none for now
    {%- endfor %}

    Methods available on node.children:
    {%- for value in methods %}
        - {{ value }}
    {%- endfor %}

    You can also iterate over node.children and access variables, render, ...
    {% raw %}
    {% for child in node.children %}
        {{ child() }}
    {% endfor %}
    {% endraw %}

    """  # noqa E501
)


class NodeTemplate(Node):
    base_path = config.path.templates

    def process(self):
        super().process()
        return self.render_template(self.template_loader(self.data.tag))

    def render_template(self, str_tpl, **kwargs):
        try:
            return render(
                "jinja",
                str_tpl,
                filters=self.get_filters(),
                **kwargs,
                node=self,
            )

        except UndefinedError as e:
            err_message = get_default_error_message(self, str_tpl)
            print(err_message)
            raise e

    def update_context(self):
        metadata = self.get_self_metadata()
        self.data.frontmatter_attributes = metadata
        super().update_context()

    def get_self_metadata(self):
        template = self.raw_template_loader(self.data.tag)
        fm_raw, template_content, handler = frontmatter_split(template)

        metadata = (
            frontmatter_parse(self.render_template(fm_raw), handler)
            if fm_raw and handler
            else {}
        )
        return metadata

    def template_loader(self, tag) -> str:
        template = self.raw_template_loader(tag)
        frontmatter, content, handler = frontmatter_split(template)
        return content.strip()

    def raw_template_loader(self, tag) -> str:
        path = self.get_template_path(tag)

        if path.exists() is False:
            self.create_default_template(path)

        return path.read_text()

    def create_default_template(self, path):
        variables = self.attributes
        methods = NodeList.get_node_list_methods()
        content = self.render_template(
            default_template_content,
            variables=variables,
            methods=list(methods),
        )
        path.write_text(content)

    def get_template_path(self, tag):
        path = get_template_path(
            self.base_path,
            tag,
            "jinja2",
            bundle=(self.attributes.get("bundle")),
        )
        return path

    @staticmethod
    def get_filters():
        return load_filters_from(config.path.filters)
